import tensorflow as tf

from .box_filter import box_filter


def guided_filter(x, y, r, eps=1e-8, nhwc=False):
    assert x.shape.ndims == 4 and y.shape.ndims == 4

    # data format
    if nhwc:
        x = tf.transpose(x, [0, 3, 1, 2])
        y = tf.transpose(y, [0, 3, 1, 2])

    # shape check
    x_shape = tf.shape(x)
    y_shape = tf.shape(y)

    assets = [tf.assert_equal(   x_shape[0],  y_shape[0]),
              tf.assert_equal(  x_shape[2:], y_shape[2:]),
              tf.assert_greater(x_shape[2:],   2 * r + 1),
              tf.Assert(tf.logical_or(tf.equal(x_shape[1], 1),
                                      tf.equal(x_shape[1], y_shape[1])), [x_shape, y_shape])]

    with tf.control_dependencies(assets):
        x = tf.identity(x)

    # N
    N = box_filter(tf.ones((1, 1, x_shape[2], x_shape[3]), dtype=x.dtype), r)

    # mean_x
    mean_x = box_filter(x, r) / N
    # mean_y
    mean_y = box_filter(y, r) / N
    # cov_xy
    cov_xy = box_filter(x * y, r) / N - mean_x * mean_y
    # var_x
    var_x  = box_filter(x * x, r) / N - mean_x * mean_x

    # A
    A = cov_xy / (var_x + eps)
    # b
    b = mean_y - A * mean_x

    mean_A = box_filter(A, r) / N
    mean_b = box_filter(b, r) / N

    output = mean_A * x + mean_b

    if nhwc:
        output = tf.transpose(output, [0, 2, 3, 1])

    return output


def fast_guided_filter(lr_x, lr_y, hr_x, r, eps=1e-8, nhwc=False):
    assert lr_x.shape.ndims == 4 and lr_y.shape.ndims == 4 and hr_x.shape.ndims == 4

    # data format
    if nhwc:
        lr_x = tf.transpose(lr_x, [0, 3, 1, 2])
        lr_y = tf.transpose(lr_y, [0, 3, 1, 2])
        hr_x = tf.transpose(hr_x, [0, 3, 1, 2])

    # shape check
    lr_x_shape = tf.shape(lr_x)
    lr_y_shape = tf.shape(lr_y)
    hr_x_shape = tf.shape(hr_x)

    assets = [tf.assert_equal(   lr_x_shape[0], lr_y_shape[0]),
              tf.assert_equal(   lr_x_shape[0], hr_x_shape[0]),
              tf.assert_equal(   lr_x_shape[1], hr_x_shape[1]),
              tf.assert_equal(  lr_x_shape[2:], lr_y_shape[2:]),
              tf.assert_greater(lr_x_shape[2:], 2 * r + 1),
              tf.Assert(tf.logical_or(tf.equal(lr_x_shape[1], 1),
                                      tf.equal(lr_x_shape[1], lr_y_shape[1])), [lr_x_shape, lr_y_shape])]

    with tf.control_dependencies(assets):
        lr_x = tf.identity(lr_x)

    # N
    N = box_filter(tf.ones((1, 1, lr_x_shape[2], lr_x_shape[3]), dtype=lr_x.dtype), r)

    # mean_x
    mean_x = box_filter(lr_x, r) / N
    # mean_y
    mean_y = box_filter(lr_y, r) / N
    # cov_xy
    cov_xy = box_filter(lr_x * lr_y, r) / N - mean_x * mean_y
    # var_x
    var_x  = box_filter(lr_x * lr_x, r) / N - mean_x * mean_x

    # A
    A = cov_xy / (var_x + eps)
    # b
    b = mean_y - A * mean_x

    # mean_A; mean_b
    A    = tf.transpose(A,    [0, 2, 3, 1])
    b    = tf.transpose(b,    [0, 2, 3, 1])
    hr_x = tf.transpose(hr_x, [0, 2, 3, 1])

    mean_A = tf.image.resize_images(A, hr_x_shape[2:])
    mean_b = tf.image.resize_images(b, hr_x_shape[2:])

    output = mean_A * hr_x + mean_b

    if not nhwc:
        output = tf.transpose(output, [0, 3, 1, 2])

    return output


import tensorflow as tf


def diff_x(input, r):
    assert input.shape.ndims == 4

    left   = input[:, :,         r:2 * r + 1]
    middle = input[:, :, 2 * r + 1:         ] - input[:, :,           :-2 * r - 1]
    right  = input[:, :,        -1:         ] - input[:, :, -2 * r - 1:    -r - 1]

    output = tf.concat([left, middle, right], axis=2)

    return output


def diff_y(input, r):
    assert input.shape.ndims == 4

    left   = input[:, :, :,         r:2 * r + 1]
    middle = input[:, :, :, 2 * r + 1:         ] - input[:, :, :,           :-2 * r - 1]
    right  = input[:, :, :,        -1:         ] - input[:, :, :, -2 * r - 1:    -r - 1]

    output = tf.concat([left, middle, right], axis=3)

    return output


def box_filter(x, r):
    assert x.shape.ndims == 4

    return diff_y(tf.cumsum(diff_x(tf.cumsum(x, axis=2), r), axis=3), r)



import time

import numpy as np
import tensorflow as tf

from skimage import img_as_float
from skimage.io import imread, imsave

from guided_filter_tf.guided_filter import guided_filter

## GuidedFilter
print('GuidedFilter:')
## check forward
# forward on img
rgb = img_as_float(imread('test/rgb.jpg'))
gt  = img_as_float(imread('test/gt.jpg'))
x, y = [tf.constant(i.transpose((2, 0, 1))[None]) for i in [rgb, gt]]
output = guided_filter(x, y, 64, 0)

with tf.Session() as sess:
    start_time = time.time()
    r = sess.run(output)
    end_time = time.time()
print('\tForward on img ...')
print('\t\tTime: {}'.format(end_time - start_time))

r = r.squeeze().transpose(1, 2, 0)
r = np.asarray(r.clip(0, 1) * 255, dtype=np.uint8)
imsave('test/r.jpg', r)


#!/usr/bin/env python
from setuptools import setup, find_packages

try:
   import pypandoc
   long_description = pypandoc.convert('README.md', 'rst')
   long_description = long_description.replace("\r", "")
except (IOError, ImportError):
    long_description = ''

setup(
    name="guided_filter_tf",
    version="1.1.1",
    description="Deep Guided Filtering Layer for TensorFlow",
    long_description=long_description,
    url="https://github.com/wuhuikai/DeepGuidedFilter",
    author="wuhuikai",
    author_email="huikaiwu@icloud.com",
    # Exclude the build files.
    packages=find_packages(exclude=["test"]),
    zip_safe=True
)


import time

import numpy as np
import tensorflow as tf

from skimage import img_as_float
from skimage.io import imread, imsave
from skimage.transform import resize

from guided_filter_tf.guided_filter import fast_guided_filter

## GuidedFilter
print('FastGuidedFilter:')
## check forward
# forward on img
rgb = img_as_float(imread('test/rgb.jpg'))
gt  = img_as_float(imread('test/gt.jpg'))
x_w, x_h = rgb.shape[:2]
w, h = x_w // 8, x_h // 8
lr_rgb = resize(rgb, (w, h), order=0, mode='reflect')
lr_gt  = resize(gt,  (w, h), order=0, mode='reflect')
lr_x, lr_y, hr_x = [tf.constant(i.transpose((2, 0, 1))[None], dtype=tf.float32) for i in [lr_rgb, lr_gt, gt]]
output = fast_guided_filter(lr_x, lr_y, hr_x, 8, 0)

with tf.Session() as sess:
    start_time = time.time()
    r = sess.run(output)
    end_time = time.time()
print('\tForward on img ...')
print('\t\tTime: {}'.format(end_time - start_time))

r = r.squeeze().transpose(1, 2, 0)
r = np.asarray(r.clip(0, 1) * 255, dtype=np.uint8)
imsave('test/r_fast.jpg', r)


import time

import numpy as np
import tensorflow as tf

from skimage import img_as_float
from skimage.io import imread

from guided_filter_tf.box_filter import box_filter

## BoxFilter
print('BoxFilter:')
# check forward
input = tf.constant(np.reshape(np.arange(1, 73), (1, 1, 8, 9)))
output = box_filter(input, 3)

with tf.Session() as sess:
    y = sess.run(output)

assert np.isclose(y.mean(),  1137.6,  0.1)
assert np.isclose(y.std(),    475.2,  0.1)
print('\tForward passed')
# forward on img
im = tf.constant(np.transpose(img_as_float(imread('test/rgb.jpg')), (2, 0, 1))[None])
output = box_filter(im, 64)
with tf.Session() as sess:
    start_time = time.time()
    r = sess.run(output)
    end_time = time.time()

print('\tForward on img ...')
print('\t\tTime: {}'.format(end_time - start_time))
r = r.squeeze().transpose(1, 2, 0)
assert np.isclose(r[:, :, 0].mean(), 10305.0, 0.1)
assert np.isclose(r[:, :, 0].std(),  2206.4,  0.1)
assert np.isclose(r[:, :, 1].mean(), 7536.0,  0.1)
assert np.isclose(r[:, :, 1].std(),  2117.0,  0.1)
assert np.isclose(r[:, :, 2].mean(), 6203.0,  0.1)
assert np.isclose(r[:, :, 2].std(),  2772.3,  0.1)
print('\tPassed ...')
